"""
    This module abstracts some of the ugliness of interfacing with C++
    from Python by providing an easy-to-use interface to the
    Neighborhood Graph Library (NGL) originally developed by Carlos
    Correa.
"""
import sklearn.neighbors

from nglpy import utils

from .ngl import nglGraph, vectorDouble, vectorInt


class EmptyRegionGraph(nglGraph):
    """A neighborhood graph that represents the connectivity of a given
    data matrix.

    Attributes:
        None
    """

    def __init__(self, *, max_neighbors=-1, relaxed=False, beta=1, p=2.0, **kwargs):
        """
        Constructor for the Empty Region Graph class that takes several keyword
        only arguments and configures a Graph object that can be applied to
        different datasets.

        This helper function will take all of the extra arguments not currently
        used by a calling function/environment, and print a warning to standard
        error letting them know they are using an unsupported feature passed into a
        function.

        Parameters
        ----------
        max_neighbors : int
            The maximum number of neighbors any one point can have. This is an
            efficiency heuristic that will limit the underlying k-nearest
            neighbor searching algorithm to only return max_neighbors edges per
            point which will then be pruned using the empty region criteria.
            This can cause a large speed-up over a full brute force edge search
            while having often little effect on the actual graph returned.
            Oftentimes, limiting the size of the empty region graph in this way
            is a benefit not a detriment for desired graph qualities.
        relaxed : bool
            Determines whether the relaxed graph (as determined by Correa and
            Lindstrom's algorithm) should be computed. A relaxed edge
            satisifies the empty region criteria from the perspective of one
            endpoint, whereas a strict edge satisfies this criteria from the
            perspective of both endpoints.
        beta : float
            A positive value working in conjunction with the p value to specify
            the size and shape of the empty region required around a valid
            edge. Roughly, beta is a size parameter and p is a shape parameter.
            For p values less than one, beta is inversely proportional to
            the size of the empty region, and for p values greater than one,
            beta is directly proportional to the size of the empty region.
            When p is one, the beta parameter is irrelevant.

            For an interactive example of how these values interact, see:
            http://www.cs.utah.edu/~maljovec/bpSkeleton.html
        p : float
            A positive value working in conjunction with the beta value to
            specify the shape and size of the empty region required around a
            valid edge. Thsi value specifies the Lp-norm to use for generating
            the empty regions. Roughly, beta is a size parameter and p is a
            shape parameter. For p values less than one, beta is inversely
            proportional to the size of the empty region, and for p values
            greater than one, beta is directly proportional to the size of the
            empty region. When p is one, the beta parameter is irrelevant.

            For an interactive example of how these values interact, see:
            http://www.cs.utah.edu/~maljovec/bpSkeleton.html

        kwargs
            Forward-compatible catch-all dictionary that will allow us to throw
            a warning for every named parameter not available in this version.

        Returns
        -------
        None

        """

        utils.consume_extra_args(kwargs)

        self.max_neighbors = max_neighbors
        self.relaxed = relaxed
        self.beta = beta
        self.p = p

    def build(self, X):
        cols = 0
        rows = len(X)
        if rows > 0:
            cols = len(X[0])

        flattened_X = [xij for Xi in X for xij in Xi]

        if self.max_neighbors <= 0:
            self.max_neighbors = rows - 1

        if self.max_neighbors >= rows:
            # Let the C++ side worry about this, do not build the knn in
            # python for a fully connected graph
            edges = vectorInt()
        else:
            knn = sklearn.neighbors.NearestNeighbors(n_neighbors=self.max_neighbors)
            knn.fit(X)
            edges = knn.kneighbors(X, return_distance=False)

            # use pairs to prevent duplicates
            pairs = []
            for e1 in range(0, edges.shape[0]):
                for col in range(0, edges.shape[1]):
                    e2 = edges.item(e1, col)
                    if e1 != e2:
                        pairs.append((e1, e2))

            # As seen here: https://bit.ly/1pUtpLh
            seen = set()
            pairs = [
                x for x in pairs if not (x in seen or x[::-1] in seen or seen.add(x))
            ]
            edgeList = []
            for edge in pairs:
                edgeList.append(edge[0])
                edgeList.append(edge[1])
            edges = vectorInt(edgeList)

        if self.relaxed:
            graph = "relaxed beta skeleton"
        else:
            graph = "beta skeleton"

        super(EmptyRegionGraph, self).__init__(
            vectorDouble(flattened_X),
            rows,
            cols,
            graph,
            self.max_neighbors,
            self.beta,
            edges,
            False,
        )

    def neighbors(self, idx=None):
        """Returns the list of neighbors associated to a particular
            index in the dataset, if one is provided, otherwise a full
            dictionary is provided relating each index to a set of
            connected indices.

        Args:
            idx: (optional) a single index of the point in the input
                data matrix for which we want to retrieve neighbors.

        Returns:
            A list of indices connected to either the provided input
            index, or a dictionary where the keys are the indices in the
            whole dataset and the values are sets of indices connected
            to the key index.
        """
        if idx is None:
            return dict(self.full_graph())
        else:
            return list(super(EmptyRegionGraph, self).get_neighbors(int(idx)))



""" This module will test the functionality of nglpy.EmptyRegionGraph
    when using the beta skeleton graph type
"""
import unittest

import nglpy


class TestBSkeleton(unittest.TestCase):
    """Class for testing the Gabriel graph"""

    def setup(self):
        """Setup function will create a fixed point set and parameter
        settings for testing different aspects of this library.

        Test graph shape:

         3----2
             /
            1
          /  |
        0     4

        However, the pruned edges should result in this graph:
        expected_graph = {0: (1, 2), 1: (0, 3, 4), 2: (0, 3, 4),
                          3: (1, 2), 4: (1, 2)}

         3----2
          | /  |
           /1  |
         //  | |
        0     4

        The relaxed version should look like:

         3----2
         |   /
        |   1
        | /  |
        0     4

        """

        self.points = [
            [0.360502, 0.535494],
            [0.476489, 0.560185],
            [0.503125, 0.601218],
            [0.462382, 0.666667],
            [0.504702, 0.5],
        ]
        self.max_neighbors = 4
        self.beta = 1
        self.relaxed = False
        self.p = 2.0
        self.discrete_steps = -1
        self.edges = [0, 1, 0, 2, 0, 3, 0, 4, 1, 3, 1, 4, 2, 3, 2, 4, 3, 4]

    def test_neighbors(self):
        """Tests the neighbors function in both settings, that is where
        an index is supplied and when it is not. This does not use
        an input neighborhood graph, thus NGL must prune the
        complete graph in this case.
        """
        self.setup()
        graph_rep = nglpy.EmptyRegionGraph(
            max_neighbors=self.max_neighbors,
            beta=self.beta,
            relaxed=self.relaxed,
            p=self.p,
            discrete_steps=self.discrete_steps,
        )
        graph_rep.build(self.points)
        expected_graph = {0: (1,), 1: (0, 2, 4), 2: (1, 3), 3: (2,), 4: (1,)}

        for i in range(len(self.points)):
            expected = list(expected_graph[i])
            actual = sorted(graph_rep.neighbors(i))
            msg = "\nNode {} Connectivity:".format(i)
            msg += "\n\texpected: {}\n\tactual: {} ".format(expected, actual)
            self.assertEqual(expected, actual, msg)

        self.assertEqual(graph_rep.neighbors(), expected_graph)

    # def test_neighbors_with_edges(self):
    #     """ Tests the neighbors function in both settings, that is where
    #         an index is supplied and when it is not. A user supplied
    #         sub-graph is used in this case. Since, this sub-graph prunes
    #         valid edges, then we should see those edges removed from the
    #         actual graph.
    #     """
    #     self.setup()
    #     # TODO: get this variable into a pre-built graph index: self.edges
    #     graph_rep = nglpy.EmptyRegionGraph(max_neighbors=self.max_neighbors,
    #                                        beta=self.beta,
    #                                        relaxed=self.relaxed,
    #                                        p=self.p,
    #                                        discrete_steps=self.discrete_steps)

    #     graph_rep.build(self.points)
    #     expected_graph = {
    #         0: (1, 2),
    #         1: (0, 3, 4),
    #         2: (0, 3, 4),
    #         3: (1, 2),
    #         4: (1, 2),
    #     }

    #     for i in range(len(self.points)):
    #         expected = list(expected_graph[i])
    #         actual = sorted(graph_rep.neighbors(i))
    #         msg = "\nNode {} Connectivity:".format(i)
    #         msg += "\n\texpected: {}\n\tactual: {} ".format(expected, actual)
    #         self.assertEqual(expected, actual, msg)

    #     self.assertEqual(graph_rep.neighbors(), expected_graph)

    def test_RelaxedNeighborhood(self):
        """Tests the neighbors function in both settings, that is where
        an index is supplied and when it is not. This does not use
        an input neighborhood graph, thus NGL must prune the
        complete graph in this case.
        """
        self.setup()
        graph_rep = nglpy.EmptyRegionGraph(
            max_neighbors=0,
            beta=self.beta,
            relaxed=True,
            p=self.p,
            discrete_steps=self.discrete_steps,
        )
        graph_rep.build(self.points)
        expected_graph = {
            0: (1, 3),
            1: (0, 2, 4),
            2: (1, 3),
            3: (0, 2),
            4: (1,),
        }

        for i in range(len(self.points)):
            expected = list(expected_graph[i])
            actual = sorted(graph_rep.neighbors(i))
            msg = "\nNode {} Connectivity:".format(i)
            msg += "\n\texpected: {}\n\tactual: {} ".format(expected, actual)
            self.assertEqual(expected, actual, msg)

        self.assertEqual(graph_rep.neighbors(), expected_graph)

    def test_empty(self):
        """Tests handling of empty data, we just want to make sure nothing
        breaks terribly.
        """
        self.setup()
        graph_rep = nglpy.EmptyRegionGraph(
            max_neighbors=self.max_neighbors,
            beta=self.beta,
            relaxed=True,
            p=self.p,
            discrete_steps=self.discrete_steps,
        )
        graph_rep.build([])
        expected_graph = {}

        self.assertEqual(graph_rep.neighbors(), expected_graph)


# TODO: Test if the kmax parameter ever gets used in this version of
# NGL, since it does not require ANN, I am assuming it does a brute
# force search of the edges if we don't provide them. We should remove
# that parameter from NGL if that is the case


if __name__ == "__main__":
    unittest.main()



""" This module will test the functionality of nglpy.PrebuiltGraph
"""
import unittest

import nglpy


class TestPrebuiltGraph(unittest.TestCase):
    """Class for testing the pre-built graph"""

    def setup(self):
        """Setup function will create a fixed point set and parameter
        settings for testing different aspects of this library.

        Test graph shape:

         3----2
             /
            1
          /  |
        0     4
        """

        self.points = [
            [0.360502, 0.535494],
            [0.476489, 0.560185],
            [0.503125, 0.601218],
            [0.462382, 0.666667],
            [0.504702, 0.5],
        ]
        self.edges = [(0, 1), (1, 2), (1, 4), (2, 3)]

    def test_neighbors(self):
        """Tests the neighbors function in both settings, that is where
        an index is supplied and when it is not. This does not use
        an input neighborhood graph, thus NGL must prune the
        complete graph in this case.
        """
        self.setup()
        graph_rep = nglpy.PrebuiltGraph(edges=self.edges)
        graph_rep.build(self.points)
        expected_graph = {0: (1,), 1: (0, 2, 4), 2: (1, 3), 3: (2,), 4: (1,)}

        for i in range(len(self.points)):
            expected = list(expected_graph[i])
            actual = sorted(graph_rep.neighbors(i))
            msg = "\nNode {} Connectivity:".format(i)
            msg += "\n\texpected: {}\n\tactual: {} ".format(expected, actual)
            self.assertEqual(expected, actual, msg)

        self.assertEqual(graph_rep.neighbors(), expected_graph)


# TODO: Test if the kmax parameter ever gets used in this version of
# NGL, since it does not require ANN, I am assuming it does a brute
# force search of the edges if we don't provide them. We should remove
# that parameter from NGL if that is the case


if __name__ == "__main__":
    unittest.main()



"""
    This module abstracts some of the ugliness of interfacing with C++
    from Python by providing an easy-to-use interface to the
    Neighborhood Graph Library (NGL) originally developed by Carlos
    Correa.
"""
from nglpy import utils

from .ngl import nglGraph, vectorDouble, vectorInt


class PrebuiltGraph(nglGraph):
    """A neighborhood graph that represents the connectivity of a given
    data matrix.

    Attributes:
        None
    """

    def __init__(self, edges=None, **kwargs):
        """
        Constructor for the Prebuilt Graph class that takes a list of edges
        and as such should only be used for one dataset.

        Parameters
        ----------
        edges : list[tuple(int,int)]
            The predefined list of edges connecting a point set

        kwargs
            Forward-compatible catch-all dictionary that will allow us to throw
            a warning for every named parameter not available in this version.

        Returns
        -------
        None

        """

        utils.consume_extra_args(kwargs)
        self.edges = edges

    def build(self, X):
        rows = len(X)
        cols = len(X[0]) if rows > 0 else 0

        flattened_X = [xij for Xi in X for xij in Xi]

        # use pairs to prevent duplicates
        # As seen here: https://bit.ly/1pUtpLh
        seen = set()
        pairs = [
            x for x in self.edges if not (x in seen or x[::-1] in seen or seen.add(x))
        ]
        edgeList = []
        for edge in pairs:
            edgeList.append(int(edge[0]))
            edgeList.append(int(edge[1]))
        edges = vectorInt(edgeList)

        super(PrebuiltGraph, self).__init__(
            vectorDouble(flattened_X),
            rows,
            cols,
            "none",
            rows,
            0,
            edges,
            False,
        )

    def neighbors(self, idx=None):
        """Returns the list of neighbors associated to a particular
            index in the dataset, if one is provided, otherwise a full
            dictionary is provided relating each index to a set of
            connected indices.

        Args:
            idx: (optional) a single index of the point in the input
                data matrix for which we want to retrieve neighbors.

        Returns:
            A list of indices connected to either the provided input
            index, or a dictionary where the keys are the indices in the
            whole dataset and the values are sets of indices connected
            to the key index.
        """
        if idx is None:
            return dict(self.full_graph())
        else:
            return list(super(PrebuiltGraph, self).get_neighbors(int(idx)))
